#include "MPU6050_6Axis_MotionApps20.h" //https://github.com/jrowberg/i2cdevlib/tree/master/Arduino/MPU6050

#define BLUE_LED_PIN A0
#define RED_LED_PIN A2
#define GREEN_LED_PIN A1
#define BASE_BALANCE_POINT_ICREMENT 0.005 // [rad]
#define MOVE_SPEED 0.1 // [m/s]

// Define increments for contants. Increments are applied on receiving characters trough bluetooth.
#define ANGLE_KP_INCREMENT 10
#define ANGLE_KI_INCREMENT 100
#define ANGLE_KD_INCREMENT 0.1
#define SPEED_KP_INCREMENT 50
#define SPEED_KI_INCREMENT 100
#define SPEED_KD_INCREMENT 0.1

#define Ext_L  2 // External interrupt left wheel signal A
#define Ext_L1  4// Encoder left wheel signal B
#define Ext_R 3 // External interrupt encoder right wheel signal A
#define Ext_R1  5 // Encoder right wheel signal B

#define WHEEL_DIAMETER 65
#define PWMENABLEPIN 12 // Used to enable pwm

#define MOTOR_ROTATION_START_OFFSET 30 // Dead zone for motors. Below this value of PWM, motors do not revolve. This constant is added to any action made by the PID.
#define PWM_MAX_CONSTRAIN 183 // Make sure max volate on motors isn't too high

// Define bluetooth protocol constants.
#define BT_LED_OFF '0'
#define BT_LED_ON '1'
#define BT_FORWARD '2'
#define BT_REVERSE '3'
#define BT_LEFT '4'
#define BT_RIGHT '5'
#define BT_PWM_STOP '6'
#define BT_PWM_START '7'
#define BT_BALANCE_UP '8'
#define BT_BALANCE_DOWN '9'
#define BT_ANGLE_FWD_KP_INC 'A'
#define BT_ANGLE_FWD_KI_INC 'B'
#define BT_ANGLE_FWD_KD_INC 'C'
#define BT_ANGLE_BCK_KP_INC 'D'
#define BT_ANGLE_BCK_KI_INC 'E'
#define BT_ANGLE_BCK_KD_INC 'F'
#define BT_SPEED_KP_INC 'G'
#define BT_SPEED_KI_INC 'H'
#define BT_SPEED_KD_INC 'I'
#define BT_ANGLE_FWD_KP_DEC 'a'
#define BT_ANGLE_FWD_KI_DEC 'b'
#define BT_ANGLE_FWD_KD_DEC 'c'
#define BT_ANGLE_BCK_KP_DEC 'd'
#define BT_ANGLE_BCK_KI_DEC 'e'
#define BT_ANGLE_BCK_KD_DEC 'f'
#define BT_SPEED_KP_DEC 'g'
#define BT_SPEED_KI_DEC 'h'
#define BT_SPEED_KD_DEC 'i'

volatile int pulsesL = 0; // LM
volatile int pulsesR = 0;

MPU6050 mpu;

// MPU control/status vars [boilerplate copied from examples in mpu library]
bool dmpReady = false;  // set true if DMP init was successful
uint8_t mpuIntStatus;   // holds actual interrupt status byte from MPU
uint8_t devStatus;      // return status after each device operation (0 = success, !0 = error)
uint16_t packetSize;    // expected DMP packet size (default is 42 bytes)
uint16_t fifoCount;     // count of all bytes currently in FIFO
uint8_t fifoBuffer[64]; // FIFO storage buffer

// orientation/motion vars
Quaternion q;           // [w, x, y, z]         quaternion container
VectorFloat gravity;    // [x, y, z]            gravity vector
float ypr[3];           // [yaw, pitch, roll]   yaw/pitch/roll container and gravity vector


const float wheel_c = PI * WHEEL_DIAMETER; // Obvod kolesa
const int counts_per_rev = 682;   // Pocet pulzov na otacku

double num_rev = (1000.0) / wheel_c;  // meter/obvod kolesa
double meter_count = num_rev * counts_per_rev; // pocet pulzov na 1 meter

double MPSspeedL = 0;//((double)pulsesL/meter_count)*1000.0/time;
double MPSspeedR = 0;//((double)pulsesR/meter_count)*1000.0/time;

// Wheel action
volatile double uL = 0;
volatile double uR = 0;

// Angle control

double angleFwdKp = 1100;// 1100;  // 1200
double angleFwdKi = 8200;// 8200; //
double angleFwdKd = 47.9;// 47.9; //
double angleFwdIntegralSum = 0;

double angleBckKp = 950; //950; //
double angleBckKi = 7000; //7000; //
double angleBckKd = 46.8; //46.8; //
double angleBckIntegralSum = 0;

double eAngle = 0;
double eOldAngle = 0;
double baseBalancePoint = -0.025; // - is analogside
double balancePointOffset = 0; // Forwads and backwards offsets
volatile double balancePoint = 0; // Calculated every loop

//  R to L speed offset regulator
double speedKp = 0;
double speedKd = 0;
double speedKi = 0;
double speedIntegral = 0;

double eSpeedOldL = 0;
double eSpeedOldR = 0;
double eSpeedL = 0;
double eSpeedR = 0;
double wantedSpeedL = 0;
double wantedSpeedR = 0;
unsigned long lastSpeedActionTime = 0;
unsigned long speedDeltaTime = 0;
int speed_action_interval_counter = 1;

// Measuring time inside loop
double deltaTime = 0; // How many ms the last loop lasted
unsigned long loopStartTime = 0;
unsigned long lastLoopStartTime = 0;

double actionValue; // Used for local calculation storage in getAction functions

bool pwmEnable = false; // Can be turend on and off with a switch, use this to turn on/off wheels (pin 12)
bool PWMEnablePinState = LOW;
bool lastPWMEnablePinState = false;
unsigned long pwmPinCooldown = 400;
unsigned long lastPWMEnableSwitchTime = 0;

int ledCounter = 0;

// Count pulses from encoders.
void intrL ()  {
  int L = digitalRead(Ext_L);
  int L1 = digitalRead(Ext_L1);

 if(L == L1){
    pulsesL++;
  }else{
    pulsesL--;
 }
}
void intrR ()  {
  int R = digitalRead(Ext_R);
  int R1 = digitalRead(Ext_R1);

  if(R == R1){
    pulsesR++;
  }else{
    pulsesR--;
  }
}

void setup() {
    Serial.begin(115200);

    // Initialize mpu device
    mpu.initialize();

    // Load and configure the DMP
    devStatus = mpu.dmpInitialize();

    // Supply your own gyro offsets here, scaled for min sensitivity
    mpu.setXGyroOffset(-45);
    mpu.setYGyroOffset(26);
    mpu.setZGyroOffset(14);
    mpu.setZAccelOffset(1490);
    mpu.setXAccelOffset(-165);
    mpu.setYAccelOffset(-879);

    // Turn on the DMP, now that it's ready
    mpu.setDMPEnabled(true);

    // Enable Arduino interrupt detection
    mpuIntStatus = mpu.getIntStatus();

    // Set our DMP Ready flag so the main loop() function knows it's okay to use it
    dmpReady = true;

    // Get expected DMP packet size for later comparison
    packetSize = mpu.dmpGetFIFOPacketSize();

    // Initialise the Motor output pins
    pinMode (6, OUTPUT);
    pinMode (7, OUTPUT);
    pinMode (8, OUTPUT);
    pinMode (9, OUTPUT);
    pinMode (10, OUTPUT);
    pinMode (11, OUTPUT);

    // Iniciatlize pushbutton pin and let pins
    pinMode (PWMENABLEPIN, INPUT);
    pinMode (RED_LED_PIN, OUTPUT);
    pinMode (BLUE_LED_PIN, OUTPUT);
    pinMode (GREEN_LED_PIN, OUTPUT);

    // By default turn off both the motors
    analogWrite(6,LOW);
    analogWrite(7,LOW);
    analogWrite(8,LOW);
    analogWrite(9,LOW);

    // Attach encoder interrupt functions.
    attachInterrupt(digitalPinToInterrupt(Ext_R), intrR, CHANGE);
    attachInterrupt(digitalPinToInterrupt(Ext_L), intrL, CHANGE);
}

void loop() {
     //// Retrieve mpu data
    mpuIntStatus = mpu.getIntStatus();

    // Get current FIFO count
    fifoCount = mpu.getFIFOCount();

    // Wait for correct available data length, should be a VERY short wait
    while (fifoCount < packetSize) fifoCount = mpu.getFIFOCount();

    // Read a packet from FIFO
    mpu.getFIFOBytes(fifoBuffer, packetSize);

    // Track FIFO count here in case there is > 1 packet available
    // (this lets us immediately read more without waiting for an interrupt)
    fifoCount -= packetSize;

    mpu.dmpGetQuaternion(&q, fifoBuffer); //get value for q
    mpu.dmpGetGravity(&gravity, &q); //get value for gravity
    mpu.dmpGetYawPitchRoll(ypr, &q, &gravity); //get value for ypr

    // Read bluetooth if availible, get the next availible char and performe the assigned action
    if (Serial.available()){
        char value = Serial.read();
        switch (value){
            case BT_LED_OFF:
                     analogWrite(RED_LED_PIN, 0);
                     analogWrite(GREEN_LED_PIN, 0);
                     analogWrite(BLUE_LED_PIN, 0);
                break;
            case BT_LED_ON: switch(ledCounter){
                        case 0:
                            analogWrite(RED_LED_PIN, 255);
                            analogWrite(GREEN_LED_PIN, 0);
                            analogWrite(BLUE_LED_PIN, 0);
                            break;
                        case 1:
                            analogWrite(RED_LED_PIN, 0);
                            analogWrite(GREEN_LED_PIN, 255);
                            analogWrite(BLUE_LED_PIN, 0);
                            break;
                        case 2:
                            analogWrite(RED_LED_PIN, 0);
                            analogWrite(GREEN_LED_PIN, 0);
                            analogWrite(BLUE_LED_PIN, 255);
                            break;
                      }
                      ledCounter = (ledCounter +1 )% 3;
                break;
            case BT_FORWARD:
                      wantedSpeedL =  MOVE_SPEED;
                      wantedSpeedR =  MOVE_SPEED;
                break;
            case BT_REVERSE:
                    wantedSpeedL =  -MOVE_SPEED;
                    wantedSpeedR =  -MOVE_SPEED;
                break;
            case BT_LEFT:
                    wantedSpeedL = 0;
                    wantedSpeedR =  MOVE_SPEED;
                break;
            case BT_RIGHT:
                    wantedSpeedR = 0;
                    wantedSpeedL =  MOVE_SPEED;
                break;
            case BT_PWM_STOP: pwmEnable = false;// Stop
                break;
            case BT_PWM_START: pwmEnable = true;// Start
                break;
            case BT_BALANCE_UP: baseBalancePoint += BASE_BALANCE_POINT_ICREMENT;// Balance_up
                break;
            case BT_BALANCE_DOWN: baseBalancePoint -= BASE_BALANCE_POINT_ICREMENT;// Balance_down
                break;
            case BT_ANGLE_FWD_KP_INC: angleFwdKp += ANGLE_KP_INCREMENT;
                break;
            case BT_ANGLE_FWD_KI_INC: angleFwdKi += ANGLE_KI_INCREMENT;
                break;
            case BT_ANGLE_FWD_KD_INC: angleFwdKd += ANGLE_KD_INCREMENT;
                break;
            case BT_ANGLE_BCK_KP_INC: angleBckKp += ANGLE_KP_INCREMENT;
                break;
            case BT_ANGLE_BCK_KI_INC: angleBckKi += ANGLE_KI_INCREMENT;
                break;
            case BT_ANGLE_BCK_KD_INC: angleBckKd += ANGLE_KD_INCREMENT;
                break;
            case BT_SPEED_KP_INC: speedKp += SPEED_KP_INCREMENT;
                break;
            case BT_SPEED_KI_INC: speedKi += SPEED_KI_INCREMENT;
                break;
            case BT_SPEED_KD_INC: speedKd += SPEED_KD_INCREMENT;
                break;
            case BT_ANGLE_FWD_KP_DEC: angleFwdKp -= ANGLE_KP_INCREMENT;
                break;
            case BT_ANGLE_FWD_KI_DEC: angleFwdKi -= ANGLE_KI_INCREMENT;
                break;
            case BT_ANGLE_FWD_KD_DEC: angleFwdKd -= ANGLE_KD_INCREMENT;
                break;
            case BT_ANGLE_BCK_KP_DEC: angleBckKp -= ANGLE_KP_INCREMENT;
                break;
            case BT_ANGLE_BCK_KI_DEC: angleBckKi -= ANGLE_KI_INCREMENT;
                break;
            case BT_ANGLE_BCK_KD_DEC: angleBckKd -= ANGLE_KD_INCREMENT;
                break;
            case BT_SPEED_KP_DEC: speedKp -= SPEED_KP_INCREMENT;
                break;
            case BT_SPEED_KI_DEC: speedKi -= SPEED_KI_INCREMENT;
                break;
            case BT_SPEED_KD_DEC: speedKd -= SPEED_KD_INCREMENT;
                break;
            default:;
        }
    }

    // Count how many ms have passed since the last loop.
    lastLoopStartTime = loopStartTime;
    loopStartTime = millis();
    deltaTime = double(loopStartTime - lastLoopStartTime);

    // Calculate setpoint
    balancePoint = baseBalancePoint + balancePointOffset;  // balancePointOffset is for forward and reverse movement, which is not implemented. baseBalancePoint is for static balacing. baseBalancePoint.
    eOldAngle = eAngle; // Save the current angle error into the oldAngle error
    eAngle = balancePoint - ypr[2]; // Calculate current error

    // Calculate pwm values for the motors using PID algorithm.
    uL = getAngleAction(eAngle, eOldAngle, deltaTime);
    uR = uL;

    // Read the pwm enable pushbutton state and change the pwmEnable global variable based on its value, if rising edge is detected.
    PWMEnablePinState = !digitalRead(PWMENABLEPIN);
    if (PWMEnablePinState == HIGH && !lastPWMEnablePinState){
       lastPWMEnablePinState = true;
       lastPWMEnableSwitchTime = millis();
       if(pwmEnable){
        pwmEnable = false;
       }else{
        pwmEnable = true;
       }
    }

    // Make sure only rising edge every pwmPinCooldown ms can trigger the change of pwmEnable global variable.
    if(millis() - lastPWMEnableSwitchTime > pwmPinCooldown && PWMEnablePinState == LOW){
      lastPWMEnablePinState = false;
    }

    // If pwmEnable is false, then do not apply the pwm values.
    if (pwmEnable){
      applyPWMToL(-uL);
      applyPWMToR(-uR);
    }else{
      applyPWMToL(0);
      applyPWMToR(0);
    }
}

// Calcuates pwm values for the motors. Uses PID algorithm. The integral part is reset whenever the controled value passes by the setpoint. Different PID settings are used for positive and negative values of e = setpoint - controled value. MOTOR_ROTATION_START_OFFSET is always added to the output, to ensure motors are not in the dead pwm zone.
double getAngleAction(double e, double eOld, double deltaTime){
    if (e*eOld < 0){
      angleBckIntegralSum = 0;
      angleFwdIntegralSum = 0;
    }

    if (e < 0){
      double propAction =  angleFwdKp*e;
      double derivAction = 1000*angleFwdKd*(e - eOld)/deltaTime;
      if (e*(e-eOld) > 0 && abs(deltaTime*angleFwdKi*angleFwdIntegralSum/1000) <255) { // integrate only till needed
        angleFwdIntegralSum += e;
      }
      double intAction = deltaTime*angleFwdKi*angleFwdIntegralSum/1000;
      actionValue = propAction+derivAction+intAction;
    }else{
      double propAction =  angleBckKp*e;
      double derivAction = 1000*angleBckKd*(e - eOld)/deltaTime;
      if (e*(e-eOld) > 0 && abs(deltaTime*angleBckKi*angleBckIntegralSum/1000) <255) { // integrate only till needed
        angleBckIntegralSum += e;
      }
      double intAction = deltaTime*angleBckKi*angleBckIntegralSum/1000;
      actionValue = propAction+derivAction+intAction;
    }
    return actionValue < 0 ? actionValue - MOTOR_ROTATION_START_OFFSET : actionValue + MOTOR_ROTATION_START_OFFSET;
}

// Applies the given pwmValue to the left weel. If the pwm value is negative, then the value is interpreted as reverse direction setting. The applied pwm value is limited by PWM_MAX_CONSTRAIN.
void applyPWMToL(int pwmValue){
    if (pwmValue > 0){ // Set forward direction for left wheel
        digitalWrite(8, HIGH);
        digitalWrite(9, LOW);
    }else{ // Set reverse direction for left wheel
        digitalWrite(8, LOW);
        digitalWrite(9, HIGH);
    }
    pwmValue = constrain(abs(pwmValue), 0, PWM_MAX_CONSTRAIN); // Apply pwm to left wheel, constrained by max pwm value.
    analogWrite(10, pwmValue);
}

// Applies the given pwmValue to the right weel. If the pwm value is negative, then the value is interpreted as reverse direction setting. The applied pwm value is limited by PWM_MAX_CONSTRAIN.
void applyPWMToR(int pwmValue){
    if (pwmValue > 0){ //  Set forward direction for right wheel
        digitalWrite(6, HIGH);
        digitalWrite(7, LOW);
    }else{ // Set reverse direction for right wheel
        digitalWrite(6, LOW);
        digitalWrite(7,HIGH);
    }
    pwmValue = constrain(abs(pwmValue), 0, PWM_MAX_CONSTRAIN); // Apply pwm to right wheel, constrained by max pwm value.
    analogWrite(11, pwmValue);
}

// Calculate wheel speeds based on encoder pulse count.
void getSpeeds(unsigned long deltaTime){

  MPSspeedL = ((double)pulsesL/meter_count)*(1000.0/(double)deltaTime);
  MPSspeedR = ((double)pulsesR/meter_count)*(1000.0/(double)deltaTime);

  pulsesL = 0;
  pulsesR = 0;

}
